@@ -21,9 +21,10 @@ for mod in list(sys.modules):
     if mod == 'josepy' or mod.startswith('josepy.'):
         sys.modules['acme.' + mod.replace('josepy', 'jose', 1)] = sys.modules[mod]
 
-if sys.version_info[0] == 2:
+
+if sys.version_info[:2] == (3, 6):
     warnings.warn(
-        "Python 2 support will be dropped in the next release of acme. "
-        "Please upgrade your Python version.",
-        PendingDeprecationWarning,
+            "Python 3.6 support will be dropped in the next release of "
+            "acme. Please upgrade your Python version.",
+            PendingDeprecationWarning,
     )  # pragma: no cover
